# -*- coding: utf-8 -*-
"""
==========================================
Orthogonal Frequency Division Multiplexing
==========================================
"""

from __future__ import annotations
from abc import ABC, abstractmethod
from enum import Enum
from math import ceil
from typing import List, Tuple, Optional, Type, Union, Any

import numpy as np
from ruamel.yaml import SafeConstructor, SafeRepresenter, MappingNode, ScalarNode
from scipy.fft import fft, fftshift, ifft, ifftshift
from scipy.interpolate import griddata
from scipy.signal import find_peaks

from ..core.factory import Serializable
from ..core.channel_state_information import ChannelStateFormat, ChannelStateInformation, ChannelStateDimension
from ..core.signal_model import Signal
from .modem import Symbols
from .waveform_generator import ChannelEqualization, PilotWaveformGenerator, Synchronization, WaveformGenerator
from .waveform_correlation_synchronization import CorrelationSynchronization
from .tools import PskQamMapping

__author__ = "André Noll Barreto"
__copyright__ = "Copyright 2022, Barkhausen Institut gGmbH"
__credits__ = ["André Noll Barreto", "Tobias Kronauer", "Jan Adler"]
__license__ = "AGPLv3"
__version__ = "0.3.0"
__maintainer__ = "Jan Adler"
__email__ = "jan.adler@barkhauseninstitut.org"
__status__ = "Prototype"


class ElementType(Enum):
    """Type of resource element."""

    REFERENCE = 0
    DATA = 1
    NULL = 2


class ChannelEstimation(Enum):
    """Applied channel estimation algorithm after reception."""

    IDEAL = 0
    IDEAL_PREAMBLE = 1
    IDEAL_MIDAMBLE = 2
    IDEAL_POSTAMBLE = 3
    REFERENCE = 4


class FrameElement:

    type: ElementType
    repetitions: int = 1

    def __init__(self,
                 type: Union[str, ElementType],
                 repetitions: int = 1) -> None:

        if type is None:
            self.type = ElementType.NULL

        elif isinstance(type, str):
            self.type = ElementType[type]

        else:
            self.type = type

        self.repetitions = repetitions


class FrameResource:
    """Configures one sub-section of an OFDM symbol section in time AND frequency."""

    __repetitions: int
    __cp_ratio: float
    elements: List[FrameElement]

    def __init__(self,
                 repetitions: int = 1,
                 cp_ratio: float = 0.0,
                 elements: Optional[List[FrameElement]] = None) -> None:

        self.repetitions = repetitions
        self.cp_ratio = cp_ratio
        self.elements = elements if elements is not None else []

    @property
    def repetitions(self) -> int:
        """Number of block repetitions along the frequency axis.

        Returns:
            int: Number of repetitions.
        """

        return self.__repetitions

    @repetitions.setter
    def repetitions(self, reps: int) -> None:
        """Modify the number of repetitions.

        Args:
            reps (int): Number of repetitions.

        Raises:
            ValueError: If `reps` is smaller than one.
        """

        if reps < 1:
            raise ValueError("Number of frame resource repetitions must be greater or equal to one")

        self.__repetitions = reps

    @property
    def cp_ratio(self) -> float:
        """Ratio between full block length and cyclic prefix.

        Returns:
            float: The ratio between zero and one.
        """

        return self.__cp_ratio

    @cp_ratio.setter
    def cp_ratio(self, ratio: float) -> None:
        """Modify the ratio between full block element length and cyclic prefix.

        Args:
            ratio (float): New ratio between zero and one.

        Raises:
            ValueError: If ratio is less than zero or larger than one.
        """

        if ratio < 0.0 or ratio > 1.0:
            raise ValueError("Cyclic prefix ratio must be between zero and one")

        self.__cp_ratio = ratio

    @property
    def num_subcarriers(self) -> int:
        """Number of occupied subcarriers.

        Returns:
            int: Number of occupied subcarriers.
        """

        num: int = 0
        for element in self.elements:
            num += element.repetitions

        return self.__repetitions * num

    @property
    def num_symbols(self) -> int:
        """Number of data symbols this resource can modulate.

        Return:
            Number of modulated symbols.
        """

        num: int = 0
        for element in self.elements:
            if element.type == ElementType.DATA:
                num += element.repetitions

        return self.__repetitions * num

    @property
    def num_references(self) -> int:
        """Number of references symbols this resource can modulate.

        Return:
            Number of modulated symbols.
        """

        num: int = 0
        for element in self.elements:
            if element.type == ElementType.REFERENCE:
                num += element.repetitions

        return self.__repetitions * num

    @property
    def mask(self) -> np.ndarray:
        """Boolean mask selecting a specific type of element from the OFDM grid.

        Returns:
            np.ndarray:
                Mask of dimension `num_element_types`x`num_subcarriers`.
        """

        # Initialize the base mask as all false
        mask = np.ndarray((len(ElementType), self.num_subcarriers), dtype=bool) * False

        element_count = 0
        for element in self.elements:

            mask[element.type.value, element_count:element_count+element.repetitions] = True
            element_count += element.repetitions

        # Repeat the subcarrier masks according to the configured number of repetitions.
        mask = np.tile(mask[:, :element_count], (1, self.__repetitions))
        return mask


class FrameSection:
    """OFDM Frame configuration time axis."""

    frame: Optional[WaveformGeneratorOfdm]
    __num_repetitions: int

    def __init__(self,
                 num_repetitions: int = 1,
                 frame: Optional[WaveformGeneratorOfdm] = None) -> None:

        self.frame = frame
        self.num_repetitions = num_repetitions

    @property
    def num_repetitions(self) -> int:
        """Number of section repetitions in the time-domain of an OFDM grid.

        Returns:
            int: The number of repetitions.
        """

        return self.__num_repetitions

    @num_repetitions.setter
    def num_repetitions(self, value: int) -> None:
        """Number of section repetitions in the time-domain of an OFDM grid.

        Args:
            value (int): The number of repetitions.

        Raises:
            ValueError: If `value` is smaller than one.
        """

        if value < 1:
            raise ValueError("OFDM frame number of repetitions must be greater or equal to one")

        self.__num_repetitions = value

    @property
    def num_symbols(self) -> int:
        """Number of data symbols this section can modulate.

        Returns:
            int: The number of symbols
        """

        return 0

    @property
    def num_references(self) -> int:
        """Number of data symbols this section can modulate.

        Returns:
            int: The number of symbols
        """

        return 0

    @property
    def num_words(self) -> int:
        """Number of OFDM symbols, i.e. words of subcarrier symbols this section can modulate.

        Returns:
            int: The number of words.
        """

        return 0

    @property
    def num_subcarriers(self) -> int:
        """Number of subcarriers this section requires.

        Returns:
            int: The number of subcarriers.
        """

        return 0

    @property
    def resource_mask(self) -> np.ndarray:
        return np.empty((len(ElementType), 0, 0), dtype=bool)

    @property
    @abstractmethod
    def num_samples(self) -> int:
        """Number of samples within this OFDM time-section.

        Returns:
            int: Number of samples
        """
        ...

    @abstractmethod
    def modulate(self, symbols: np.ndarray) -> np.ndarray:
        """Modulate this section into a complex base-band signal.

        Args:
            symbols (np.ndarray):
                The complex data symbols encoded in this OFDM section.

        Returns:
            np.ndarray: The modulated signal vector.
        """
        ...

    @abstractmethod
    def demodulate(self,
                   signal: np.ndarray,
                   channel_state: ChannelStateInformation) -> Tuple[np.ndarray, ChannelStateInformation]:
        """Demodulate a time section of a complex OFDM base-band signal into data symbols.

        Args:
            signal (np.ndarray): Vector of complex-valued base-band samples.
            channel_state (ChannelStateInformation): Channel state.

        Returns:
            (np.ndarray, channel_state):
                Section symbol grid and channel response grid.
        """
        ...


class FrameSymbolSection(FrameSection, Serializable):

    yaml_tag: str = u'Symbol'
    pattern: List[int]

    def __init__(self,
                 num_repetitions: int = 1,
                 pattern: Optional[List[int]] = None,
                 frame: Optional[WaveformGeneratorOfdm] = None) -> None:

        FrameSection.__init__(self, num_repetitions=num_repetitions, frame=frame)
        self.pattern = pattern if pattern is not None else []
        self.frame = frame

    @property
    def num_symbols(self) -> int:

        num = 0
        for resource_idx in self.pattern:

            resource = self.frame.resources[resource_idx]
            num += resource.num_symbols

        return self.num_repetitions * num

    @property
    def num_references(self) -> int:

        num = 0
        for resource_idx in self.pattern:

            resource = self.frame.resources[resource_idx]
            num += resource.num_references

        return self.num_repetitions * num

    @property
    def num_words(self) -> int:
        return self.num_repetitions * len(self.pattern)

    @property
    def num_subcarriers(self) -> int:

        num = 0

        for resource_idx in set(self.pattern):
            num = max(num, self.frame.resources[resource_idx].num_subcarriers)

        return num

    def modulate(self, symbols: np.ndarray) -> np.ndarray:

        # Collect resource masks
        mask = self.resource_mask[:, :self.num_subcarriers]

        # Generate the resource grid of the oversampled OFDM frame
        padded_num_subcarriers = self.frame.num_subcarriers * self.frame.oversampling_factor
        grid = np.zeros((padded_num_subcarriers, self.num_words), dtype=complex)
        
        # Select the subgrid onto which to project this symbol section's resource configuration
        subgrid_start_idx = int(.5 * (padded_num_subcarriers - self.num_subcarriers))
        subgrid = grid[subgrid_start_idx:subgrid_start_idx+self.num_subcarriers, :]

        # Reference fields all currently carry the complex symbol 1+j0
        # ToDo: Implement reference symbol configurations
        subgrid[mask[ElementType.REFERENCE.value]] = 1. + 0j

        # Data fields carry the supplied data symbols
        subgrid.T[mask[ElementType.DATA.value].T] = symbols

        # NULL fields are just that... zero
        subgrid[mask[ElementType.NULL.value]] = 0j
        
        # Shift in order to suppress the dc component
        # Note that for configurations without any oversampling the DC component will not be suppressed
        if self.frame.dc_suppression:
            
            dc_index = int(.5 * padded_num_subcarriers)
            grid[dc_index:, :] = np.roll(grid[dc_index:, :], 1, axis=0)

        # By convention, the length of each time slot is the inverse of the sub-carrier spacing
        resource_signals = ifft(ifftshift(grid, axes=0), axis=0, norm='ortho')

        # Add the cyclic prefix to each time slot while simultaneously flatten the resource signals into time domain
        signals = []
        for resource_idx, resource_samples in enumerate(resource_signals.T):

            pattern_idx = resource_idx % len(self.pattern)
            cp_ratio = self.frame.resources[self.pattern[pattern_idx]].cp_ratio

            num_prefix_samples = int(padded_num_subcarriers * cp_ratio)

            if num_prefix_samples > 0:
                signals.append(resource_samples[-num_prefix_samples:])

            signals.append(resource_samples)

        signal_samples = np.concatenate(signals, axis=0)
        return signal_samples

    def demodulate(self,
                   signal: np.ndarray,
                   channel_state: ChannelStateInformation) -> Tuple[np.ndarray, ChannelStateInformation]:

        padded_num_subcarriers = self.frame.num_subcarriers * self.frame.oversampling_factor

        # Remove the cyclic prefixes before transformation into time-domain
        sample_index = 0
        sample_indices = np.empty(0, dtype=int)
        channel_sample_indices = np.empty(0, dtype=int)
        num_slots = len(self.pattern) * self.num_repetitions

        for slot_idx in range(num_slots):

            pattern_idx = slot_idx % len(self.pattern)
            resource = self.frame.resources[self.pattern[pattern_idx]]

            num_prefix_samples = int(padded_num_subcarriers * resource.cp_ratio)
            sample_index += num_prefix_samples

            sample_indices = np.append(sample_indices, np.arange(sample_index, sample_index + padded_num_subcarriers))
            channel_sample_indices = np.append(channel_sample_indices, np.array(sample_index))

            sample_index += padded_num_subcarriers

        slot_samples = signal[sample_indices].reshape((padded_num_subcarriers, num_slots), order='F')
        slot_channel_state = channel_state[:, :, channel_sample_indices, :]\
            .to_frequency_selectivity(num_bins=self.frame.num_subcarriers)

        # Transform grid back to data symbols
        grid = fftshift(fft(slot_samples, axis=0, norm='ortho'), axes=0)

        # Account for the DC suppression
        if self.frame.dc_suppression:

            dc_index = int(.5 * padded_num_subcarriers)
            grid[dc_index:, :] = np.roll(grid[dc_index:, :], -1, axis=0)
            
        # Extract the subgrid relevant to this section
        subgrid_start_idx = int(.5 * (padded_num_subcarriers - self.num_subcarriers))
        subgrid = grid[subgrid_start_idx:subgrid_start_idx+self.num_subcarriers, :]

        return subgrid, slot_channel_state

    @property
    def resource_mask(self) -> np.ndarray:

        # Initialize the base mask as all false
        mask = np.zeros((len(ElementType), self.num_subcarriers, len(self.pattern)), dtype=bool)

        for word_idx, resource_idx in enumerate(self.pattern):

            resource = self.frame.resources[resource_idx]
            mask[:, :resource.num_subcarriers, word_idx] = resource.mask

        return np.tile(mask, (1, 1, self.num_repetitions))

    @property
    def num_samples(self) -> int:

        num_samples_per_slot = self.frame.num_subcarriers * self.frame.oversampling_factor
        num = len(self.pattern) * num_samples_per_slot

        # Add up the additional samples from cyclic prefixes
        for resource_idx in self.pattern:
            num += int(num_samples_per_slot * self.frame.resources[resource_idx].cp_ratio)

        # Add up the base samples from each timeslot
        return num * self.num_repetitions

    @classmethod
    def from_yaml(cls: Type[FrameSymbolSection],
                  constructor: SafeConstructor,
                  node: Union[ScalarNode, MappingNode]) -> FrameSymbolSection:

        if isinstance(node, ScalarNode):
            return cls()

        return cls(**constructor.construct_mapping(node))

    @classmethod
    def to_yaml(cls: Type[FrameSymbolSection], representer: SafeRepresenter, node: FrameSymbolSection) -> MappingNode:

        state = {
            'num_repetitions': node.num_repetitions,
            'pattern': node.pattern,
        }

        return representer.represent_mapping(node.yaml_tag, state)


class FrameGuardSection(FrameSection, Serializable):

    yaml_tag: str = u'Guard'
    __duration: float

    def __init__(self,
                 duration: float,
                 num_repetitions: int = 1,
                 frame: Optional[WaveformGeneratorOfdm] = None) -> None:

        FrameSection.__init__(self, num_repetitions=num_repetitions, frame=frame)
        self.duration = duration

    @property
    def duration(self) -> float:
        """Guard section duration in seconds.

        Returns:
            float: Duration in seconds.
        """

        return self.__duration

    @duration.setter
    def duration(self, value: float) -> None:
        """Guard section duration in seconds.

        Args:
            value (float): New duration.

        Raises:
            ValueError: If `value` is smaller than zero.
        """

        if value < 0.0:
            raise ValueError("Guard section duration must be greater or equal to zero")

        self.__duration = value

    @property
    def num_samples(self) -> int:

        return int(self.num_repetitions * self.__duration * self.frame.sampling_rate)

    def modulate(self, symbols: np.ndarray) -> np.ndarray:

        if len(symbols) > 0:
            raise ValueError("Guard sections may not hold modulation symbols")

        return np.zeros(self.num_samples, dtype=complex)

    def demodulate(self,
                   baseband_signal: np.ndarray,
                   channel_state: ChannelStateInformation) -> Tuple[np.ndarray, ChannelStateInformation]:

        # Guard sections naturally don't encode anything
        return np.empty((self.frame.num_subcarriers, 0), dtype=complex), ChannelStateInformation.Ideal(0)

    @classmethod
    def from_yaml(cls: Type[FrameGuardSection],
                  constructor: SafeConstructor,
                  node: MappingNode) -> FrameGuardSection:

        return cls(**constructor.construct_mapping(node))

    @classmethod
    def to_yaml(cls: Type[FrameGuardSection], representer: SafeRepresenter, node: FrameGuardSection) -> MappingNode:

        state = {
            'num_repetitions': node.num_repetitions,
            'duration': node.duration,
        }

        return representer.represent_mapping(cls.yaml_tag, state)


class WaveformGeneratorOfdm(PilotWaveformGenerator, Serializable):
    """Generic Orthogonal-Frequency-Division-Multiplexing with a flexible frame configuration.

    The following features are supported:
        - The modem can transmit or receive custom-defined frames.
          Frames may contain UL/DL data symbols, null carriers, pilot subcarriers,
          reference signals and guard intervals.
        - SC-FDMA can also be implemented with a precoder.
        - Subcarriers can be modulated with BPSK/QPSK/16-/64-/256-QAM.
        - Cyclic prefixes for interference-free channel estimation and equalization are supported.

    This implementation has currently the following limitations:
        - All subcarriers use the same modulation scheme


    Attributes:

        __channel_estimation_algorithm (ChannelEstimation):
            Method deployed to simulate OFDM channel estimation.

        __num_subcarriers (int:
            Maximum number of subcarriers.
            Also the size of the FFT deployed during modulation,
            i.e. the difference between the configured number of subcarriers and the maximum number
            will be zero-padded.

        dc_suppression (bool):
            Suppress the direct current component during waveform generation.

        resources (List[FrameResource]):
            Frequency-domain resource section configurations.

        structure (List[FrameSection]):
            Time-domain frame configuration.
    """

    yaml_tag: str = u'OFDM'

    __channel_estimation_algorithm: ChannelEstimation
    __subcarrier_spacing: float
    __num_subcarriers: int
    __pilot_section: Optional[FrameSection]
    dc_suppression: bool
    resources: List[FrameResource]
    structure: List[FrameSection]

    def __init__(self,
                 channel_estimation: Union[str, ChannelEstimation] = ChannelEstimation.IDEAL,
                 subcarrier_spacing: float = 1e3,
                 num_subcarriers: int = 1200,
                 dc_suppression: bool = True,
                 resources: Optional[List[FrameResource]] = None,
                 structure: Optional[List[FrameSection]] = None,
                 **kwargs: Any) -> None:
        """Orthogonal-Frequency-Division-Multiplexing Waveform Generator initialization.

        Args:

            channel_estimation (Union[str, ChannelEstimation], optional):
                Method deployed to simulate OFDM channel estimation.

            subcarrier_spacing (float, optional):
                Spacing between individual subcarriers in Hz.

            num_subcarriers (int, optional):
                Maximum number of subcarriers.
                Also the size of the FFT deployed during modulation,
                i.e. the difference between the configured number of subcarriers and the maximum number
                will be zero-padded.

            dc_suppression (bool, optional):
                Suppress the direct current component during waveform generation.

            resources (List[FrameResource], optional):
                Frequency-domain resource section configurations.

            structure (List[FrameSection], optional):
                Time-domain frame configuration.

            kwargs (Any):
                Waveform generator base class initialization parameters.
        """

        # Init base class
        PilotWaveformGenerator.__init__(self, **kwargs)

        self.channel_estimation_algorithm = channel_estimation
        self.subcarrier_spacing = subcarrier_spacing
        self.num_subcarriers = num_subcarriers
        self.dc_suppression = dc_suppression
        self.resources = [] if resources is None else resources
        self.channel_equalization = OFDMChannelEqualization(self,)
        self.__pilot_section = None

        self.structure = []
        if structure is not None:
            for section in structure:
                self.add_section(section)

        self._mapping = PskQamMapping(self.modulation_order)

    def add_resource(self, resource: FrameResource) -> None:
        """Add a OFDM frequency resource to the waveform.

        Args:
            resource (FrameResource):
                The resource description to be added.
        """

        self.resources.append(resource)

    def add_section(self, section: FrameSection) -> None:
        """Add a frame section to the OFDM structure.

        Args:

            section (FrameSection):
                The section to be added.
        """

        self.structure.append(section)
        section.frame = self
        
    @property
    def pilot_section(self) -> Optional[FrameSection]:
        """Static pilot section transmitted at the beginning of each OFDM frame.
        
        Required for time-domain synchronization and equalization of carrier frequency offsets.
        
        Returns:
            FrameSection of the pilot symbols, `None` if no pilot is configured.
        """
        
        return self.__pilot_section
    
    @pilot_section.setter
    def pilot_section(self, value: Optional[FrameSection]) -> None:
        
        if value is None:
            self.__pilot_section = None
            return
            
        self.__pilot_section = value
        
        if value.frame is not self:
            value.frame = self
            
    @property
    def pilot_signal(self) -> Signal:
        
        if self.pilot_section:
            return Signal(self.pilot_section.modulate(), sampling_rate=self.sampling_rate)

        else:
            return Signal.empty(self.sampling_rate)

    @property
    def channel_estimation_algorithm(self) -> ChannelEstimation:
        """Used channel estimation algorithm.

        Return:
            ChannelEstimation:
                The channel estimation algorithm.
        """

        return self.__channel_estimation_algorithm

    @channel_estimation_algorithm.setter
    def channel_estimation_algorithm(self, value: Union[str, ChannelEstimation]) -> None:
        """Modify the used channel estimation algorithm.

        Args:
            value (Union[str, ChannelEstimation]):
                New channel estimation algorithm.
        """

        if isinstance(value, str):
            self.__channel_estimation_algorithm = ChannelEstimation[value]

        else:
            self.__channel_estimation_algorithm = value

    @property
    def subcarrier_spacing(self) -> float:
        """Subcarrier spacing between frames.

        Returns:
            float: Spacing in Hz.
        """

        return self.__subcarrier_spacing

    @subcarrier_spacing.setter
    def subcarrier_spacing(self, spacing: float) -> None:
        """Modify the subcarrier spacing between frames.

        Args:
            spacing (float): New spacing in Hz.

        Raises:
            ValueError: If `spacing` is smaller or equal to zero.
        """

        if spacing <= 0.0:
            raise ValueError("Subcarrier spacing must be greater than zero")

        self.__subcarrier_spacing = spacing

    @property
    def symbols_per_frame(self) -> int:

        num_symbols = 0
        for section in self.structure:
            num_symbols += section.num_symbols

        return num_symbols

    @property
    def words_per_frame(self) -> int:

        num_words = 0
        for section in self.structure:
            num_words += section.num_words

        return num_words

    @property
    def references_per_frame(self) -> int:

        num_symbols = 0
        for section in self.structure:
            num_symbols += section.num_references

        return num_symbols

    @property
    def frame_duration(self) -> float:
        
        return self.samples_in_frame / self.sampling_rate

    @property
    def samples_in_frame(self) -> int:
        """int: Returns read-only samples_in_frame"""

        num = 0
        for section in self.structure:
            num += section.num_samples
            
        if self.pilot_signal:
            num += self.pilot_signal.num_samples

        return num

    def map(self, data_bits: np.ndarray) -> Symbols:
        return Symbols(self._mapping.get_symbols(data_bits))

    def unmap(self, data_symbols: Symbols) -> np.ndarray:

        detected_bits = self._mapping.detect_bits(data_symbols.raw).astype(int)
        return detected_bits

    def modulate(self, data_symbols: Symbols) -> Signal:
       
        # Start the frame with a pilot section, if configured
        if self.pilot_section:
            output_signal = self.pilot_section.modulate()

        else:
            output_signal = np.empty(0, dtype=complex)

        # Convert symbols
        data_symbols = data_symbols.raw.flatten()

        sent_data_symbols = 0
        for section in self.structure:

            # Number of data symbols encoded in this time-section of the OFDM frame
            section_num_data_symbols = section.num_symbols

            # Select the data symbols to be sent over this time-section of the OFDM frame
            section_data_symbols = data_symbols[sent_data_symbols:sent_data_symbols+section_num_data_symbols]
            sent_data_symbols += section_num_data_symbols

            # Modulate the signal
            section_signal = section.modulate(section_data_symbols)
            output_signal = np.append(output_signal, section_signal)

        signal_model = Signal(output_signal, self.sampling_rate)
        return signal_model

    def demodulate(self,
                   signal: np.ndarray,
                   channel_state: ChannelStateInformation,
                   noise_variance: float = 0.0) -> Tuple[Symbols, ChannelStateInformation, np.ndarray]:

        # Recover OFDM grid
        symbol_grid = np.empty((self.num_subcarriers, self.words_per_frame), dtype=complex)
        resource_mask = np.zeros((len(ElementType), self.num_subcarriers, self.words_per_frame), dtype=bool)
        section_channel_states: List[ChannelStateInformation] = []

        sample_index = 0
        word_index = 0
        
        # If the frame contains a pilot section, skip the respective samples
        if self.pilot_section:
            sample_index += self.pilot_section.num_samples
        
        for section in self.structure:

            num_samples = section.num_samples
            num_words = section.num_words

            if num_words < 1:

                sample_index += num_samples
                continue

            time_indices = np.arange(sample_index, sample_index+num_samples)
            signal_section = signal[time_indices]
            channel_state_section = channel_state[:, :, time_indices, :]

            section_symbol_grid, section_channel_state = section.demodulate(signal_section, channel_state_section)
            section_mask = section.resource_mask

            symbol_grid[:section_mask.shape[1], word_index:word_index+num_words] = section_symbol_grid
            section_channel_states.append(section_channel_state)
            resource_mask[:, :section_mask.shape[1], word_index:word_index+num_words] = section_mask

            sample_index += num_samples
            word_index += num_words

        ideal_channel_state = ChannelStateInformation.concatenate(section_channel_states,
                                                                  dimension=ChannelStateDimension.SAMPLES)

        # Estimate the channel given the recovered OFDM resources and convert it back to linear transformation matrices
        # Since we handle frequency bins here, the CSI transformations are diagonal over the last two dimensions
        channel_state_estimation = self.__channel_estimation(symbol_grid, ideal_channel_state, resource_mask)

        # Recover the data symbols, as well as the respective channel weights from the resource grids
        data_mask = resource_mask[ElementType.DATA.value]
        channel_state_estimation = channel_state_estimation[:, :, data_mask.flatten(), :]
        data_symbols = symbol_grid.T[data_mask.T]
        noise_variances = np.repeat(noise_variance, self.symbols_per_frame)

        return Symbols(data_symbols), channel_state_estimation, noise_variances

    @property
    def bandwidth(self) -> float:

        # OFDM bandwidth currently is identical to the number of subcarriers times the subcarrier spacing
        b = self.num_subcarriers * self.subcarrier_spacing
        return b

    def  __channel_estimation(self,
                             symbol_grid: np.ndarray,
                             channel_state: ChannelStateInformation,
                             resource_mask: np.ndarray) -> ChannelStateInformation:
        """Performs channel estimation over the OFDM grid.

        This methods estimates the frequency response of the channel for all OFDM symbols in a frame. The estimation
        algorithm is defined in the parameter variable `self.param`.

        With ideal channel estimation, the channel state information is obtained directly from the ideal channel state.
        The CSI can be considered to be known only at the beginning/middle/end of the frame
        (estimation_type='IDEAL_PREAMBLE'/'IDEAL_MIDAMBLE'/ 'IDEAL_POSTAMBLE'), or at every OFDM symbol ('IDEAL').

        With reference-based estimation, the specified reference subcarriers are employed for channel estimation.

        Args:

            symbol_grid (numpy.ndarray):
                Frequency-domain samples of the received signal over the whole frame.

            channel_state (ChannelStateInformation):
                Perfect channel state from which to simulate the channel state estimation.

            resource_mask (np.ndarray):
                Boolean mask for OFDM resource allocation.
                Required to distinguish between data, reference and null symbols within `symbol_grid`.

        Returns:
            ChannelStateInformation:
                The channel state estimate resulting from the selected method.
        """

        # Ideally, the channel is estimated perfectly at each received symbol slot
        if self.channel_estimation_algorithm == ChannelEstimation.IDEAL:
            return channel_state

        # Number of modulation symbols per ofdm word
        num_symbols = symbol_grid.shape[0]
        num_words = symbol_grid.shape[1]

        # An ideal pre-amble estimates the channel at the first sample position
        if self.channel_estimation_algorithm == ChannelEstimation.IDEAL_PREAMBLE:

            estimate = channel_state.state[:, :, :num_symbols, :]
            channel_state.state = np.tile(estimate, (1, 1, num_words, 1))
            return channel_state

        # An ideal mid-amble estimates the channel at the central symbol position
        if self.channel_estimation_algorithm == ChannelEstimation.IDEAL_MIDAMBLE:

            word_idx = int(.5 * num_words)
            estimate = channel_state.state[:, :, word_idx*num_symbols:(1+word_idx)*num_symbols, :]
            channel_state.state = np.tile(estimate, (1, 1, num_words, 1))
            return channel_state

        # An ideal post-amble estimates the channel at the last sample position
        if self.channel_estimation_algorithm == ChannelEstimation.IDEAL_POSTAMBLE:

            estimate = channel_state.state[:, :, -num_symbols:, :]
            channel_state.state = np.tile(estimate, (1, 1, num_words, 1))
            return channel_state

        if self.channel_estimation_algorithm == ChannelEstimation.REFERENCE:
            return self.reference_based_channel_estimation(symbol_grid, resource_mask)

        raise RuntimeError("Unknown OFDM channel estimation routine requested")

    @staticmethod
    def reference_based_channel_estimation(symbol_grid: np.ndarray,
                                           resource_mask: np.ndarray) -> ChannelStateInformation:
        """Perform a reference-symbol based channel estimation over the OFDM frame grid.

        This method estimates the channel using reference symbols. Only LS method is currently implemented. The function
        will return only a single value for each subcarrier. If several reference symbols are available, then the
        estimate will be averaged over all OFDM symbols.

        Args:

            symbol_grid (numpy.ndarray):
                Frequency-domain samples of the received signal over the whole frame.

            resource_mask (np.ndarray):
                Boolean mask for OFDM resource allocation.
                Required to distinguish between data, reference and null symbols within `symbol_grid`.

        Returns:

            ChannelStateInformation:
                The channel state estimate.
        """

        propagated_reference_symbols = symbol_grid.T[resource_mask[ElementType.REFERENCE.value, ::].T]
        reference_symbols = np.ones(len(propagated_reference_symbols), dtype=complex)
        reference_channel_estimation = propagated_reference_symbols / reference_symbols

        channel_estimation = np.zeros(symbol_grid.shape, dtype=complex)
        channel_estimation.T[resource_mask[ElementType.REFERENCE.value, ::].T] = reference_channel_estimation

        interpolation_stems = np.where(resource_mask[ElementType.REFERENCE.value, ::])
        holes = np.where(np.invert(resource_mask[ElementType.REFERENCE.value, ::]))

        # ToDo: Check with group what to do about missing values outside the convex hull
        interpolated_holes = griddata(interpolation_stems, reference_channel_estimation, holes, method='nearest')
        channel_estimation[holes] = interpolated_holes
        return ChannelStateInformation(ChannelStateFormat.FREQUENCY_SELECTIVITY, channel_estimation.T.flatten()[None, None, :, None])


    @property
    def bits_per_frame(self) -> int:
        return self.symbols_per_frame * self._mapping.bits_per_symbol

    @property
    def bit_energy(self) -> float:

        return 1 / self._mapping.bits_per_symbol  # ToDo: Re-implement
        # return self.oversampling_factor / self._mapping.bits_per_symbol * self._cyclic_prefix_overhead

    @property
    def symbol_energy(self) -> float:

        return 1 / self._mapping.bits_per_symbol  # ToDo: Re-implement
        # return self.oversampling_factor * self._cyclic_prefix_overhead

    @property
    def power(self) -> float:

        return 1  # ToDo: Re-implement
#        return self.num_occupied_subcarriers / self.fft_size

    @property
    def num_subcarriers(self) -> int:
        """Maximum number of subcarriers.

        Sometimes also referred to as FFT-size.

        Returns:
            int: Number of subcarriers.
        """

        return self.__num_subcarriers

    @num_subcarriers.setter
    def num_subcarriers(self, value: int) -> None:
        """Modify the maximum number of subcarriers.

        Args:
            value (int): New maximum number of subcarriers.

        Raises:
            ValueError: If `value` is smaller than one.
        """

        if value < 1:
            raise ValueError("Number of subcarriers must be greater or equal to one")

        self.__num_subcarriers = value

    @property
    def sampling_rate(self) -> float:
        return self.oversampling_factor * self.subcarrier_spacing * self.__num_subcarriers

    @classmethod
    def to_yaml(cls: Type[WaveformGeneratorOfdm],
                representer: SafeRepresenter,
                node: WaveformGeneratorOfdm) -> MappingNode:
        """Serialize an `WaveformGenerator` object to YAML.

        Args:
            representer (SafeRepresenter):
                A handle to a representer used to generate valid YAML code.
                The representer gets passed down the serialization tree to each node.

            node (WaveformGeneratorOfdm):
                The `WaveformGeneratorOfdm` instance to be serialized.

        Returns:
            Node:
                The serialized YAML node
        """

        state = {
            'channel_estimation': node.__channel_estimation_algorithm.value,
            'subcarrier_spacing': node.__subcarrier_spacing,
            'num_subcarriers': node.__num_subcarriers,
            'dc_suppression': node.dc_suppression
        }

        mapping = representer.represent_mapping(cls.yaml_tag, state)
        mapping.value.extend(WaveformGenerator.to_yaml(representer, node).value)

        return mapping

    @classmethod
    def from_yaml(cls: Type[WaveformGeneratorOfdm], constructor: SafeConstructor, node: MappingNode)\
            -> WaveformGeneratorOfdm:
        """Recall a new `WaveformGeneratorOfdm` instance from YAML.

        Args:
            constructor (SafeConstructor):
                A handle to the constructor extracting the YAML information.

            node (Node):
                YAML node representing the `WaveformGeneratorOfdm` serialization.

        Returns:
            WaveformGeneratorOfdm:
                Newly created `WaveformGeneratorOfdm` instance.
        """

        state = constructor.construct_mapping(node, deep=True)
        state = {k.lower(): v for k, v in state.items()}

        structure: List[FrameSection] = state.pop('structure', None)
        resources = state.pop('resources', None)

        # Handle resource list to object conversion
        if resources is not None:
            for resource_idx, resource in enumerate(resources):

                element_objects = []
                elements = resource.pop('elements', [])
                for element_args in elements:
                    element_objects.append(FrameElement(**element_args))
                resource['elements'] = element_objects

                resources[resource_idx] = FrameResource(**resource)

        state['resources'] = resources

        # Create actual frame object from state dictionary
        ofdm = cls(**state)

        if structure is not None:
            for section in structure:
                ofdm.add_section(section)

        return ofdm
    

class PilotSection(FrameSection):
    """Pilot symbol section within an OFDM frame."""
    
    __pilot_elements: Optional[Symbols]
    __cached_num_subcarriers: int
    __cached_oversampling_factor: int
    __cached_pilot: Optional[np.ndarray]

    
    def __init__(self,
                 pilot_elements: Optional[Symbols] = None,
                 frame: Optional[WaveformGeneratorOfdm] = None) -> None:
        """
        Args:
        
            pilot_elements (Optional[Symbols], optional):
                Symbols with which the subcarriers within the pilot will be modulated.
                By default, a pseudo-random sequence from the frame mapping will be generated.

            frame (Optional[WaveformGeneratorOfdm], optional):
                The frame configuration this pilot section belongs to.
        """
        
        self.__pilot_elements = pilot_elements
        self.__cached_num_subcarriers = -1
        self.__cached_oversampling_factor = -1
        self.__cached_pilot = None
        
        FrameSection.__init__(self, num_repetitions=1, frame=frame)
        
    @property
    def num_samples(self) -> int:
        
        return self.frame.num_subcarriers * self.frame.oversampling_factor
        
    @property
    def pilot_elements(self) -> Optional[Symbols]:
        """Symbols with which the subcarriers within the pilot will be modulated.
        
        Returns:
        
            A stream of symbols. `None`, if no subsymbols where specified.
            
        Raises:
        
            ValueError: If the configured symbols contains multiple streams.
        """
        
        return self.__pilot_elements
    
    @pilot_elements.setter
    def pilot_elements(self, value: Optional[Symbols]) -> None:
        
        if value is None:
            self.__pilot_elements = None
            return
        
        if value.num_streams != 1:
            raise ValueError("Subsymbol pilot configuration may only contain a single stream")
        
        if value.num_symbols < 1:
            raise ValueError("Subsymbol pilot configuration must contain at least one symbol")
        
        # Reset the cached pilot, since the subsymbols have changed
        self.__cached_pilot = None
        
        self.__pilot_elements = value
            
    def _pilot_sequence(self, num_symbols: int = None) -> Symbols:
        """Generate a new sequence of pilot elements.
        
        Args:
        
            num_symbols (int, optional):
                The required number of symbols.
                By default, a symbol for each subcarrier is generated.
                
        Returns:
        
            A sequence of symbols.
        """
        
        num_symbols = self.frame.num_subcarriers if num_symbols is None else num_symbols
        
        # Generate a pseudo-random symbol stream if no subsymbols are specified
        if self.__pilot_elements is None:
            
            rng = np.random.default_rng(42)
            num_bits = num_symbols * self.frame._mapping.bits_per_symbol
            subsymbols = self.frame._mapping.get_symbols(rng.integers(0, 2, num_bits))[None, :]
            
        else:
            
            num_repetitions = int(ceil(num_symbols / self.__pilot_elements.num_symbols))
            subsymbols = np.tile(self.__pilot_elements.raw, (1, num_repetitions))
            
        return Symbols(subsymbols[:, :num_symbols])
    
    def modulate(self, *_: Any) -> np.ndarray:
        
        # Return the cached pilot signal if available and the relevant frame parameters haven't changed
        if self.__cached_pilot is not None and self.__cached_num_subcarriers == self.frame.num_subcarriers and self.__cached_oversampling_factor == self.frame.oversampling_factor:
            return self.__cached_pilot
        
        pilot = self._pilot()
        self.__cached_pilot = pilot
        
        return pilot
    
    def demodulate(self, *_: Any) -> Tuple[np.ndarray, ChannelStateInformation]:
        
        return np.empty(0, dtype=complex), ChannelStateInformation(ChannelStateFormat.FREQUENCY_SELECTIVITY)
        
    def _pilot(self) -> np.ndarray:
        """Generate the samples for a pilot section in time domain.
        
        Returns:
        
            Complex base-band pilot section samples.
        """
        
        samples_per_symbol = self.frame.num_subcarriers * self.frame.oversampling_factor
        pilot = ifft(self._pilot_sequence().raw[0, :], norm='ortho', n=samples_per_symbol)
        return pilot


class SchmidlCoxPilotSection(PilotSection):
    """Pilot Symbol Section of the Schmidl Cox Algorithm.
    
    Refer to :footcite:t:`1997:schmidl` for a detailed description.
    """

    @property
    def num_samples(self) -> int:
        
        return 2 * self.frame.num_subcarriers * self.frame.oversampling_factor

    def _pilot(self) -> np.ndarray:

        samples_per_symbol = self.frame.num_subcarriers * self.frame.oversampling_factor
        num_symbols = int(1.5 * self.frame.num_subcarriers)
        pilot_sequence = self._pilot_sequence(num_symbols).raw.flatten()

        pilot_frequencies = np.zeros((self.frame.num_subcarriers, 2), dtype=complex)
        pilot_frequencies[0::2, 0] = pilot_sequence[self.frame.num_subcarriers:]
        pilot_frequencies[:, 1] = pilot_sequence[:self.frame.num_subcarriers]
 
        pilot_symbols_time = ifft(pilot_frequencies, axis=0, norm='ortho', n=samples_per_symbol)
        pilot_samples = np.concatenate(pilot_symbols_time.T, axis=0)
        
        return pilot_samples
    
    def demodulate(self, *_: Any) -> Tuple[np.ndarray, ChannelStateInformation]:
        
        return np.empty(0, dtype=complex), ChannelStateInformation(ChannelStateFormat.FREQUENCY_SELECTIVITY)
    

class OFDMSynchronization(Synchronization[WaveformGeneratorOfdm]):
    """Synchronization Routine for OFDM Waveforms."""
    ...


class OFDMCorrelationSynchronization(CorrelationSynchronization[WaveformGeneratorOfdm]):
    """Correlation-Based Pilot Detection and Synchronization for OFDM Waveforms."""
    ...


class SchmidlCoxSynchronization(OFDMSynchronization):
    """Schmidl-Cox Algorithm for OFDM Waveform Time Synchronization and Carrier Frequency Offset Equzalization.
    
    Applying the synchronization routine requires the respective waveform to have a :class:`.SchmidlCoxPilotSection` pilot
    symbol section configured.

    Refer to :footcite:t:`1997:schmidl` for a detailed description.
    """

    yaml_tag = u'SchmidlCox'
    """YAML serialization tag"""

    def synchronize(self,
                    signal: np.ndarray,
                    channel_state: ChannelStateInformation) -> List[Tuple[np.ndarray, ChannelStateInformation]]:

        symbol_length = self.waveform_generator.oversampling_factor * self.waveform_generator.num_subcarriers
        
        # Abort if the supplied signal is shorter than one symbol length
        if signal.shape[-1] < symbol_length:
            return []
            
        half_symbol_length = int(.5 * symbol_length)

        num_delay_candidates = signal.shape[-1] - symbol_length
        delay_powers = np.empty(1 + num_delay_candidates, dtype=float)
        delay_powers[0] = 0.
        for d in range(0, num_delay_candidates):

            delay_powers[1 + d] = np.sum(abs(np.sum(signal[:, d:d + half_symbol_length].conj() * signal[:, d + half_symbol_length:d + 2 * half_symbol_length], axis=1)))
        
        num_samples = self.waveform_generator.samples_in_frame
        peaks, _ = find_peaks(delay_powers, distance=int(.8 * num_samples))
        frame_indices = peaks - 1

        frames: List[Tuple[np.ndarray, ChannelStateInformation]] = []
        for frame_idx in frame_indices:

            frames.append((signal[:, frame_idx:frame_idx + num_samples], channel_state[:, :, frame_idx:frame_idx + num_samples, :]))

        return frames
    

class OFDMChannelEqualization(ChannelEqualization[WaveformGeneratorOfdm], ABC):
    """Channel estimation for OFDM waveforms."""

    def __init__(self,
                 waveform_generator: Optional[WaveformGeneratorOfdm] = None) -> None:
        """
        Args:

            waveform_generator (WaveformGenerator, optional):
                The waveform generator this equalization routine is attached to.
        """

        ChannelEqualization.__init__(self, waveform_generator)


class OFDMZeroForcingChannelEqualization(Serializable, OFDMChannelEqualization, ABC):
    """Zero-Forcing Channel estimation for Psk Qam waveforms."""
    
    yaml_tag = u'OFDM-ZF'
    """YAML serialization tag"""

    def __init__(self,
                 waveform_generator: Optional[WaveformGeneratorOfdm] = None) -> None:
        """
        Args:

            waveform_generator (WaveformGenerator, optional):
                The waveform generator this equalization routine is attached to.
        """

        OFDMChannelEqualization.__init__(self, waveform_generator)

    def equalize_channel(self,
                         frame: np.ndarray,
                         csi: ChannelStateInformation,
                         snr: float = float('inf')) -> Signal:

        
        equalized_frame = frame / csi.state[0, 0, :, 0]
        return equalized_frame


class OFDMMinimumMeanSquareChannelEqualization(Serializable, OFDMChannelEqualization, ABC):
    """Minimum-Mean-Square Channel estimation for Psk Qam waveforms."""
    
    yaml_tag = u'OFDM-MMSE'
    """YAML serialization tag"""

    def __init__(self,
                 waveform_generator: Optional[WaveformGeneratorOfdm] = None) -> None:
        """
        Args:

            waveform_generator (WaveformGenerator, optional):
                The waveform generator this equalization routine is attached to.
        """

        OFDMChannelEqualization.__init__(self, waveform_generator)

    def equalize_channel(self,
                         signal: Signal,
                         csi: ChannelStateInformation,
                         snr: float = float('inf')) -> Signal:

        signal = signal.copy()
        signal.samples /= (csi.state[0, 0, :signal.num_samples, 0] + 1 / snr)
        
        return signal
